eval 'exec perl -w -S $0 ${1+"$@"}'
    if 0;

################################################################################
#
# $RCSfile: httpremote_amarok.pl $
#
#
# This program is Copyright (C) 03/2006 Matthias Prinke
# <matthias.prinke@surfeu.de> and covered by GNU's GPL.
# In particular, this program is free software and comes WITHOUT
# ANY WARRANTY.
#
# $Author: mp $
#
# $Date: 2006/12/27 $
#
# $Revision: 0.0.8 $
#
################################################################################

use strict;
use warnings;

use HTTP::Daemon;
use HTTP::Response;
use HTTP::Headers;
use HTTP::Status;
use XML::Simple;
use IO::Select;
use DCOP::Amarok::Player;
use DCOP::Amarok::Playlist;
use MIME::Base64::Perl;
# The Tk-Modules are loaded by installed() if available
#use Tk;
#use Tk::NumEntry;

our $VERSION    = '0.0.8';
my $inetaddr    = (defined($ENV{HOSTNAME})) ? inet_aton("$ENV{HOSTNAME}")
                                            : undef;
my $ipaddress   = (defined($inetaddr)) ? inet_ntoa($inetaddr) : "127.0.0.1";
my %config;

#### BEGIN CONFIG
my $user                = "$ENV{USER}"; # username
my $verbose             = 0;            # verbosity flag
$config{'tcp_port'}     = 50000;        # TCP port to use (in private range)
$config{'autorefresh'}  = 0;            # auto refresh - not supported by
                                        #   some browsers, fails completely
                                        #   on Palm Web Pro V3.0
$config{'refresh_rate'} = 5;            # update rate in seconds
				        #   for autorefresh	  
$config{'host'}         = $ipaddress;   # host name for auto refresh

                                        # Context page replacement options:
$config{'replace_data'}         = 1;    # replace inline images by files
$config{'replace_file'}         = 1;    # replace file: links by http request
$config{'replace_musicbrainz'}  = 1;    # replace by link to control page
$config{'replace_fetchcover'}   = 1;    # replace by link to control page
$config{'replace_togglebox'}    = 1;    # remove togglebox: href
$config{'replace_album'}        = 1;    # remove album: href
#### END   CONFIG

=head1 NAME

httpremote_amarok

=head1 DESCRIPTION

This program serves as a remote control / status display
based on a web browser. The browser may run on the same
machine as amaroK, on another desktop machine or - which
makes most sense - on a smart phone/PDA/web tablet/...
connected via a wireless network interface (Bluetooth/WLAN/IrDA).

Technically, the script provides a HTTP server listening on
a private TCP port and sends a dynamically generated HTML
page with some amarok status information and some links
to the client. By selecting the links in the browser, the
user can send various commands to the server which are passed
to amaroK. 

=cut

my $cfgfile   = "httpremote_config.xml";  # configuration file
my @mod_check = qw(Tk Tk::NumEntry);      # optional perl modules
my $have_tk   = 1;                        # flag: Tk is available
my $c;                                    # socket connection handle

# amaroK Player Object
my $amarok    = DCOP::Amarok::Player->new( user => "$user" )
                or die "Couldn't attach to DCOP: $!\n";

# amaroK Playlist Object
my $amarok_pl = DCOP::Amarok::Playlist->new( user => "$user" )
                or die "Couldn't attach to DCOP: $!\n";

# Select Object
my $readset   = new IO::Select();


# script's path to find the logo image file
# (the amaroK-scripts are started from the directory
# ~/.kde/share/apps/amarok/scripts-data while the scripts are in
# ~/.kde/share/apps/amarok/scripts)
$0 =~ /(\S+)\/httpremote_amarok.pl/;
my $img_path  = $1;


# read configuration if available
if (-r $cfgfile) {
  my $config_r = XMLin($cfgfile);
  %config = %$config_r;
} else {
  # write initial config file
  write_config($cfgfile, \%config);
}


# Check if Perl Modules for Configuration GUI are available
for (@mod_check) {
  if ( not(installed("$_")) ) {
    print "httpremote_amarok: $_ is not installed.\n";
    print "httpremote_amarok: Configuration GUI will not be available.\n";
    $have_tk = 0;
    last;
  }
}


# Create HTTP server.
# The option ReuseAddr is used to allow dangling tcp connections (in TIME_WAIT
# state) to be reused immediatly.
my $d = HTTP::Daemon->new(
  LocalPort => $config{'tcp_port'},
  ReuseAddr => 1
) || die;
print "Please contact me at: <URL:", $d->url, "httpremote>\n" if $verbose;


# catch TERM signal
$SIG{TERM} = \&bye;

$readset->add($d, \*STDIN);

while(1) {
  # select(): read from socket or STDIN, wherever input is available
  if (my @ready = $readset->can_read) {  
    foreach my $handle (@ready) {
  
      if ($handle == $d) {
        # read from socket
	
	$c = $d->accept;
        my $r = $c->get_request;

	if (($r->method eq 'GET') and ($r->url->path eq "/httpremote")) {
	  if ($r->url =~ /cmd=play$/) {
	    print "Command: play\n" if $verbose;
	    $amarok->play();
	  
	  } elsif ($r->url =~ /\?cmd=stop/) {
	    print "Command: stop\n" if $verbose;
	    $amarok->stop();
	  
	  } elsif ($r->url =~ /\?cmd=pause/) {
	    print "Command: pause\n" if $verbose;
	    $amarok->pause();
	  
	  } elsif ($r->url =~ /\?cmd=fwd/) {
	    print "Command: fwd\n" if $verbose;
	    $amarok->fwd();
	  
	  } elsif ($r->url =~ /\?cmd=rew/) {
	    print "Command: rew\n" if $verbose;
	    $amarok->rew();
	  
	  } elsif ($r->url =~ /\?cmd=next/) {
	    print "Command: next\n" if $verbose;
	    $amarok->next();
	  
	  } elsif ($r->url =~ /\?cmd=prev/) {
	    print "Command: prev\n" if $verbose;
	    $amarok->prev();
	  
	  } elsif ($r->url =~ /\?cmd=vol_up/) {
	    print "Command: vol+\n" if $verbose;
	    $amarok->volUp();
	  
	  } elsif ($r->url =~ /\?cmd=vol_dn/) {
	    print "Command: vol-\n" if $verbose;
	    $amarok->volDn();
	  
	  } elsif ($r->url =~ /\?cmd=mute/) {
	    print "Command: mute\n" if $verbose;
	    $amarok->mute();
	  
	  } elsif ($r->url =~ /\?cmd=rnd/) {
	    print "Command: mute\n" if $verbose;
	    $amarok->toggleRandom();
	  
	  } elsif ($r->url =~ /\?cmd=play_index&index=(\d+)/) {
	    print "Command: playByIndex($1)\n" if $verbose;
	    $amarok_pl->playByIndex($1);
	  
	  } elsif ($r->url =~ /\?cmd=play_media&path=(.+)/) {
            print "Command: playMedia($1)\n" if $verbose;
	    my $url = $1;
	    
	    # modify spaces in URL string
	    $url =~ s/%20/ /g;
	    
	    $amarok_pl->playMedia("\"$url\"");
	  
	  }
	  
	  my $header  = HTTP::Headers->new(Content_Type => "text/html");
	  my $content;
	  
	  if (($r->url =~ /\?cmd=playlist$/)   ||
	      ($r->url =~ /\?cmd=play_index/)) {
	    $content = playlist_page();
	    
	  } elsif (($r->url =~ /\?cmd=context$/)     ||
	           ($r->url =~ /\?cmd=play_media/)) {
	    $content = context_page();
	    
	  } else {
	    $content = main_page()
	    
	  }

	  my $resp = HTTP::Response->new(
	    RC_OK,
	    status_message(RC_OK),
	    $header,
	    $content
	  );
	  $c->send_response($resp);
	  $header = undef;
	  $resp = undef;

	} elsif (($r->method eq 'GET') and ($r->url->path eq "/logo.png")) {
	  # return the amarok logo
	  $c->send_file_response("$img_path/logo.png");

	} elsif (($r->method eq 'GET') and ($r->url =~ /\.png/)) {
	  # return an arbitrary .png image
	  if (-r $r->url) {
	    # absolute path specified
	    # i.e. <host>:<port>/usr/local/share/data/mp3/<myalbum>/cover.png
	    $c->send_file_response($r->url);
	  
	  } else {
	    # relative path specified
	    # i.e. <host>:<port>/context_img<n>.png
	    # strip the leading slash from the path
	    my $url = substr($r->url, 1);
	    if (-r $url) {
	      $c->send_file_response($url);
	    }
	  
	  }
	
	} else {
            $c->send_error(RC_FORBIDDEN)
        }
	$c->close;
        undef($c);

      } elsif ($handle == \*STDIN) {
        $_ = <STDIN>;
      
        if ((/configure/) && ($have_tk == 1)) {
	  print "\nhttpremote_amarok: configure\n" if $verbose;
	  config_gui();
	  
	} elsif ( /trackChange/ ) {
          printf("\n%s-%s\n", $amarok->artist(), $amarok->title())if $verbose
	} elsif ( /engineStateChange/ ) {
	  if( /playing/ ) {
	    printf("\nhttpremote_amarok: Playing %s\n",$amarok->title()) if $verbose;
	  } elsif ( /pause/ ) {
	    print "\nhttpremote_amarok: pause\n" if $verbose;
	  } else {
	    print "\nhttpremote_amarok: stopped\n" if $verbose;
	  }
	}
	exit if(/exit/); #for debugging
	&bye() if(/kill/); # first kill, then exit... after thread has exited        
      } # elsif ($handle == STDIN)
    } # foreach my $handle (@ready) {
  } # if (my @ready = $readset->can_read)
} # while (1)


# generate main HTML page with some dynamic content
sub main_page
{
  my $resp;
  my $volume  = $amarok->vol();
  my $artist  = $amarok->artist();
  my $album   = $amarok->album();
  my $title   = $amarok->title();
  my $track   = $amarok->track();
  my $elapsed = $amarok->elapsed();
  my $total   = $amarok->totaltime();
  my $random  = ($amarok->getRandom() eq "true") ? "RND" : "";
  my $rnd_btn = ($amarok->getRandom() eq "true") ? "RND Off" : "RND On";
  my $refresh = ($config{'autorefresh'}) ? 
  "<meta http-equiv=\"refresh\" content=\"$config{'refresh_rate'}; \
  URL=http://$config{'host'}:$config{'tcp_port'}/httpremote\">" : "";
    
  $resp = <<END_OF_TEXT;
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
<head>
<title>amaroK</title>
$refresh
<style type="text/css">
A{text-decoration:none; color:blue}
</style>
</head>
<body>
<center><a href="httpremote"><img src="logo.png" alt="amaroK" border="0"></a></center>
<center>
<p>
<table border="1">
<tr>
<td>Volume</td>
<td>$volume</td>
</tr>
<tr>
<td>Track</td>
<td>$track $random</td>
</tr>
<tr>
<td>Artist</td>
<td>$artist</td>
</tr>
<tr>
<td>Album</td>
<td>$album</td>
</tr>
<tr>
<td>Title</td>
<td>$title</td>
</tr>
<tr>
<td>Time</td>
<td>$elapsed/$total</td>
</tr>
</table>
</p>
<p>
<table border="1">
<tr>
<td><a href="/httpremote?cmd=play">Play</a></td>
<td><a href="/httpremote?cmd=pause">Pause</a></td>
<td><a href="/httpremote?cmd=stop">Stop</a></td>
<td><a href="/httpremote?cmd=rnd">$rnd_btn</a></td>
<td><a href="/httpremote?cmd=prev">&lt; Prev</a></td>
<td><a href="/httpremote?cmd=next">Next &gt;</a></td>
</tr>
</table>
</p>
<p>
<table border="1">
<tr>
<td><a href="/httpremote?cmd=context">Context</a></td>
<td><a href="/httpremote?cmd=playlist">Playlist</a></td>
<td><a href="/httpremote?cmd=rew">&lt;&lt; Rew</a></td>
<td><a href="/httpremote?cmd=fwd">Fwd &gt;&gt;</a></td>
</tr>
</table>
</p>
<p>
<table border="1">
<tr>
<td><a href="/httpremote?cmd=vol_dn">Vol-</a></td>
<td><a href="/httpremote?cmd=vol_up">Vol+</a></td>
<td><a href="/httpremote?cmd=mute">Mute</a></td>
</tr>
</table>
</p>
</center>
</body>
</html>
END_OF_TEXT

  return $resp;
} # sub main_page


# show playlist
sub playlist_page
{
  my $resp = "";
  my $index = 0;
  my $path = $amarok_pl->saveCurrentPlaylist();
  my $ok = not($@);  
  chomp($path);
#  my $playlist = XMLin("/tmp/current.xml", forcearray => [ 'item' ]);
  my $playlist = XMLin("$path", forcearray => [ 'item' ]);
  my $current = $amarok_pl->getActiveIndex();
  chomp($current);
  my $curr_link = "";
  
  if ($current > -1) {
    $curr_link = "<br>&gt; <a href=\"#current\">Currently Playing</a> &lt;";
  } 

  # Note:
  # Since the links contain only the index into the playlist,
  # marking links already selected is not meaningful when a
  # playlist is modified, so we override the link color to
  # a fixed value. Shame on browsers which do not support it.
  
  $resp .= <<END_OF_HEAD;
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
<head>
<title>amaroK</title>
<meta http-equiv="Content-type" content="text/html; charset=iso-8859-1" />
<style type="text/css">
A{color:blue}
</style>
</head>
<body>
<h1>amaroK Playlist</h1>
&lt;&lt; <a href="/httpremote">Back</a> $curr_link
<p></p>
<table>
END_OF_HEAD
 
  if ($ok) {
    foreach my $item (@{$playlist->{item}}) {
      last if ( !defined($item) );

      my $tag_on   = ($index == $current) ? "<b>" : "";
      my $tag_off  = ($index == $current) ? "</b>" : "";
      my $tag_name = ($index == $current) ? "<a name=\"current\">" : "";
      my $title    = ( defined($item->{Title} ) ) ? $item->{Title}  : "";
      my $artist   = ( defined($item->{Artist}) ) ? $item->{Artist} : "";

      $resp .= "<tr>\n";
      $resp .= "<td>$tag_name<a href=\"/httpremote?cmd=play_index&index=$index\">"; 
      $resp .= $tag_on . $title . $tag_off . "</a></td>\n";
      $resp .= "<td><a href=\"/httpremote?cmd=play_index&index=$index\">";
      $resp .= $tag_on . $artist . $tag_off . "</a></td>\n";
      $resp .= "</tr>\n";
      $index++;
    } # foreach my $item (@{$playlist->{item}})
  } # if ($ok)
 
  $resp .= <<END_OF_TAIL;
</table>
<p>
&lt;&lt; <a href="/httpremote">Back</a>
</p>
</body>
</html>
END_OF_TAIL
  
  return $resp;
} # sub playlist_page


# Perl Module availability checking subroutine
sub installed {
  my $module = $_;

  my $check = 0;

  eval "use $module";
  if ($@) {
    $check = 0;
  } else {
    $check = 1; # non-zero equals installed.
    my $version = 0;
    eval "\$version = \$$module\::VERSION";
    $check = $version if (!$@); # Set check to version number if it exists
  }
} # sub installed


# Create configuration GUI
sub config_gui
{
  my $mw = MainWindow->new();
  $mw->resizable(0, 0);
  $mw->title( "httpremote" );

  # Host Name/IP Address Entry
  $mw->Label(-text => "Host Name/IP Address:", -justify => 'left')
     	   ->pack(-side => 'top', -anchor => 'w');

  my $host = $mw->Entry(-justify => 'left',
     	   -textvariable => \$config{'host'})
     	   ->pack(-fill => 'x', -side => 'top');

  # empty frame as spacer
  $mw->Frame()
     	   ->pack(-fill => 'x', -side => 'top', -pady => 5);
  
  # TCP Port Entry
  $mw->Label(-text => "TCP Port (49152..65535):", -justify => 'left')
     	   ->pack(-side => 'top', -anchor => 'w');

  my $port = $mw->NumEntry(-justify => 'left', -validate => 'key',
     	   -minvalue => 49152, -maxvalue => 65535,
     	   -value => $config{'tcp_port'})
     	   ->pack(-fill => 'x', -side => 'top');

  # empty frame as spacer
  $mw->Frame()
     	   ->pack(-fill => 'x', -side => 'top', -pady => 5);

  # Auto Refresh Checkbutton
  my $refresh = $mw->Checkbutton(-text => "Auto Refresh",
     	   -variable => \$config{'autorefresh'})
     	   ->pack(-side => 'top', -anchor => 'w');

  # Refresh Rate Entry
  $mw->Label(-text => "Refresh Rate (sec):", -justify => 'left')
     	   ->pack(-side => 'top', -anchor => 'w');

  my $rate = $mw->NumEntry(-justify => 'left', -validate => 'key',
     	   -minvalue => 1, -value => $config{'refresh_rate'})
     	   ->pack(-fill => 'x', -side => 'top');

  # empty frame as spacer
  $mw->Frame()
     	   ->pack(-fill => 'x', -side => 'top', -pady => 5);

  # Context Page Modifications
  $mw->Label(-text => "Context Page Modifications:", -justify => 'left')
     	   ->pack(-side => 'top', -anchor => 'w', -ipady => 2);

  $mw->Checkbutton(-text => "Replace Inline Images",
     	   -variable => \$config{'replace_data'})
     	   ->pack(-side => 'top', -anchor => 'w');

  $mw->Checkbutton(-text => "Replace File References",
     	   -variable => \$config{'replace_file'})
     	   ->pack(-side => 'top', -anchor => 'w');

  $mw->Checkbutton(-text => "Replace 'musicbrainz:' Reference",
     	   -variable => \$config{'replace_musicbrainz'})
     	   ->pack(-side => 'top', -anchor => 'w');

  $mw->Checkbutton(-text => "Replace 'fetchcover:' References",
     	   -variable => \$config{'replace_fetchcover'})
     	   ->pack(-side => 'top', -anchor => 'w');

  $mw->Checkbutton(-text => "Replace 'togglebox:' Reference",
     	   -variable => \$config{'replace_togglebox'})
     	   ->pack(-side => 'top', -anchor => 'w');

  $mw->Checkbutton(-text => "Replace 'album:' References",
     	   -variable => \$config{'replace_album'})
     	   ->pack(-side => 'top', -anchor => 'w');

  # hostname and refresh rate only valid if autorefresh is enabled
  if ($config{'autorefresh'} eq "0") {
    $host->configure(-state => 'disabled');
    $rate->configure(-state => 'disabled');
  }

  # configure callback for autorefresh checkbutton
  $refresh->configure(-command => sub {
     	   if ($config{'autorefresh'} eq "1") {
     	     $host->configure(-state => 'normal');
     	     $rate->configure(-state => 'normal');
     	   } else {
     	     $host->configure(-state => 'disabled');
     	     $rate->configure(-state => 'disabled');
     	     $refresh->focus;
     	   }
  });

  # empty frame as spacer
  $mw->Frame()
     	   ->pack(-fill => 'x', -side => 'top', -pady => 5);

  # Cancel Button
  $mw->Button(-text => "Cancel",
     	   -command => sub {
     	     # Cancel - ignore all input
     	     $mw->destroy;
     	     $mw = undef})
     	   ->pack(-fill => 'x', -side => 'left');
  
  # OK Button
  $mw->Button(-text => "OK", -default => 'active', -takefocus => '1',
     	   -padx => 25,
     	   -command => sub {
     	     # OK - read and process input
     	     if ($port->get() != $config{'tcp_port'}) {
     	       # port changed - restart server
     	       $d = undef;
     	       $readset = undef;
     	       $config{'tcp_port'}     = $port->get();
     	       $d = HTTP::Daemon->new(
		      LocalPort => $config{'tcp_port'},
     		      ReuseAddr => 1
     		    ) || die;

     	       # socket descriptor might have changed -
     	       #   reconfigure select
     	       $readset = new IO::Select();
     	       $readset->add($d, \*STDIN);

     	     }
     	     $config{'refresh_rate'} = $rate->get();

     	     # write configuration to file
     	     write_config($cfgfile, \%config);
     	     $mw->destroy;
     	     $mw = undef})
     	  ->pack(-fill => 'x',-side => 'right');

  &MainLoop;
} # sub config_gui


# write config file
sub write_config {
  my $cfgfile  = shift; 
  my $config_r = shift;
  my $xml = XMLout($config_r);
  
  if (open (CFGFILE, ">$cfgfile")) {
    print CFGFILE $xml;
    close(CFGFILE);
  }
} # sub write_config


# read and convert contextbrowser page
sub context_page {
  my $slurp = "";
  my $cnt = -1;

  if (open (INFILE, "$img_path/../../contextbrowser.html")) {
    while ($_ = <INFILE>) {
      $slurp .= $_;
    }
    close(INFILE);
    

    # extract inline images and replace by links
    # Inline images like <img src='data:image/png;base64,...'> are not supported
    # by some browsers, including a popular product by a Redmont-based software
    # company...
    #   match "src='..." globally, execute extract_img on each match,
    #   replace matching string by return value
    if ($config{'replace_data'} == 1) {
      $slurp =~ s/src='data:image\/png;base64,([^']+)'/ &extract_img($1, \$cnt) /ge;
    }
    
    # replace file-protocol by command-URL
    # FIXME: this should be done only for media file links 
    if ($config{'replace_file'} == 1) {
      $slurp =~ s/file:/\/httpremote?cmd=play_media&path=/g;
    }
    
    # replace unknown protocol links for images by "back to main page"
    if ($config{'replace_musicbrainz'} == 1) {
      $slurp =~ s/'musicbrainz:([^']+)'/'\/httpremote'/;
    }
    if ($config{'replace_fetchcover'} == 1) {
      $slurp =~ s/'fetchcover:([^']+)'/'\/httpremote'/g;
    }
    
    # remove 'togglebox:' reference
    if ($config{'replace_togglebox'} == 1) {
      # window.location.href='togglebox:ft';
      $slurp =~ s/window\.location\.href='togglebox:ft';//g;
    }
    
    # replace
    # <a href='album:405 @@@ 308'><span class='album-title'>Title</span></a>
    # by
    # <span class='album-title'>Title</span>
    if ($config{'replace_album'} == 1) {
      $slurp =~ s/<a href='album:[^']+'>(.*?)<\/a>/$1/g;
    }
  }
  
  return $slurp;
} # sub context_page


# extract, decode and write image data to file
# in ~/.kde/share/apps/amarok/scripts-data/,
# return partial image source link string  
sub extract_img {
  my $data  = shift;
  my $cnt_r = shift;
  
  $$cnt_r++;
  
  open (IMGFILE, ">context_img$${cnt_r}.png");
  print IMGFILE decode_base64($data);
  close(IMGFILE);
  
  return "src='context_img$${cnt_r}.png'";
} # extract_img


# clean up before leaving
sub bye {
  # client object
  $c->close;
  undef($c);
  
  # daemon object
  undef($d);
  
  exit;
} # sub bye

__END__

=head1 AUTHOR

M. Prinke E<lt>matthias.prinke@surfeu.deE<gt>

=head1 COPYRIGHT (COPYLEFT)

2006, M. Prinke

=head1 LICENSE

This programs is licensed under the terms of the GPL. 

=cut
